/***************************************************************************
 *   Copyright (C) 2006 by Parthasarathy   *
 *   sarathy@Zion   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

# include <math.h>
# include <stdio.h>
# include <stdlib.h>

# include <constants.h>
# include <algorithms.h>
# include <output.h>

/*
        Method to sort the given data. The unsorted data, the
        algorithm to be used and the sorting order must be specified.
        This is the only externally visible method in this file.
*/
void sort_data (long **unsorted_data, long n, int algorithm, int ascending)
{
	switch (algorithm)
	{
		case INSERTION_SORT:
			insertion_sort(unsorted_data, n, ascending);
			break;
		
		case INSERTION_RECURSIVE_SORT:
			insertion_recursive_sort(unsorted_data, n, ascending);
			break;
		
		case MERGE_SORT:
			merge_sort(unsorted_data, 0,n-1, ascending);
			break;
	
		case SELECTION_SORT:
			selection_sort(unsorted_data, n, ascending);
			break;

		case BUBBLE_SORT:
			bubble_sort(unsorted_data, n, ascending);
			break;

		case QUICK_SORT:
			quick_sort(unsorted_data, 0, n-1, ascending);
			break;
		default:
			break;
	}
}

/* Sort data using insertion sort.*/
static void insertion_sort(long **unsorted_data_ptr, long n, int ascending)
{
	long i=0,j=0;
	long key=0;
	long *unsorted_data = *unsorted_data_ptr;


	for (j=1;j<n;j++)
	{
		key = *(unsorted_data+j);
		i=j-1;

		if (ascending ^ 1)
		{
			while (i>=0  && *(unsorted_data+i) >key)
			{
				*(unsorted_data+i+1)=*(unsorted_data+i);
				i--;
			}
		}
		else
		{
			while (i>=0  && *(unsorted_data+i) <key)
			{
				*(unsorted_data+i+1)=*(unsorted_data+i);
				i--;
			}
		}
		*(unsorted_data+i+1)=key;
	}
}

/* Sort data using selection sort.*/
static void selection_sort (long **unsorted_data_ptr, long n, int ascending)
{
	long i=0;
	long start=0;
	long smallest=0;
	long temp_data=0;
	long *unsorted_data=*unsorted_data_ptr;
	
	while (start<n-1)
	{
		smallest = start;

		if (ascending ^ 1)
		{
			for (i=start;i<n-1;i++)
			{
				if ( *(unsorted_data+i+1) < *(unsorted_data+smallest))
					smallest=i+1;
			}
		}
		else
		{
			for (i=start;i<n-1;i++)
			{
				if ( *(unsorted_data+i+1) > *(unsorted_data+smallest))
					smallest=i+1;
			}
		}
		temp_data = *(unsorted_data + start);
		*(unsorted_data+start)=*(unsorted_data+smallest);
		*(unsorted_data+smallest)=temp_data;
		start++;
	}
}

/* Sort data using merge sort.*/
void merge_sort(long **unsorted_data, long start, long end, int ascending)
{
	if (start<end)
	{
		merge_sort(unsorted_data,start,start+(end-start)/2, ascending);
		merge_sort(unsorted_data,start+(end-start)/2+1,end,ascending);
		merge(unsorted_data,start,start+(end-start)/2,end, ascending);
	}
	else
		return;

}

/* Recursive merge of data.*/
static void merge (long **unsorted_data_ptr, long idx_start, long idx_middle, long idx_end, int ascending)
{
	long i=0;

	long idx_subseq1=0;
	long idx_subseq2=0;

	long n1=idx_middle-idx_start+1;
	long n2=idx_end-idx_middle;
	
	long *unsorted_data = *unsorted_data_ptr;
	long *subseq1=(long*)malloc((n1+1)*sizeof(long));
	long *subseq2=(long*)malloc((n2+1)*sizeof(long));

	for (i=idx_start;i<=idx_middle;i++)
		*(subseq1+i-idx_start)=*(unsorted_data+i);
	for (i=idx_middle+1;i<=idx_end;i++)
		*(subseq2+i-idx_middle-1)=*(unsorted_data+i);

	/* 
	   Two for loops have been used to increase performance, since a single for loop will force comparison at each step to
	   check for ascending or descending order decreasing the performance.
	 */
	if (ascending ^ 1)
	{
		for (i=idx_start;i<=idx_end;i++)
		{
			if (*(subseq1+idx_subseq1)<*(subseq2+idx_subseq2))
			{
				*(unsorted_data+i)=*(subseq1+idx_subseq1);
				idx_subseq1++;
			}
			else
			{
				*(unsorted_data+i)=*(subseq2+idx_subseq2);
				idx_subseq2++;
			}
		
			if (idx_subseq1==n1 || idx_subseq2==n2)
				break;
		}
	} 
	else 
	{
		for (i=idx_start;i<=idx_end;i++)
		{
			if (*(subseq1+idx_subseq1)>*(subseq2+idx_subseq2))
			{
				*(unsorted_data+i)=*(subseq1+idx_subseq1);
				idx_subseq1++;
			}
			else
			{
				*(unsorted_data+i)=*(subseq2+idx_subseq2);
				idx_subseq2++;
			}
		
			if (idx_subseq1==n1 || idx_subseq2==n2)
				break;
		}
	}

	if (idx_subseq1 == n1)
	{
		while (idx_subseq2!=n2)
		{
			i++;
			*(unsorted_data+i)=*(subseq2+idx_subseq2);
			idx_subseq2++;
		}
	}


	if (idx_subseq2 == n2)
	{
		while (idx_subseq1!=n1)
		{
			i++;
			*(unsorted_data+i)=*(subseq1+idx_subseq1);
			idx_subseq1++;
		}
	}
}

/* Sort data using recursive insertion sort.*/
static void insertion_recursive_sort(long **unsorted_data_ptr, long n, int ascending)
{
	long key=0;
	long j=0;
	long *unsorted_data = *unsorted_data_ptr;

	if (n==1)
		return;
	else
	{
		insertion_recursive_sort(unsorted_data_ptr,n-1, ascending);
		key = *(unsorted_data+n-1);
		
		if (ascending ^ 1) 
			for (j=n-2;j>=0 && key<*(unsorted_data+j);j--)
			       	*(unsorted_data+j+1)=*(unsorted_data+j);
		else
			for (j=n-2;j>=0 && key>*(unsorted_data+j);j--)
				*(unsorted_data+j+1)=*(unsorted_data+j);
		*(unsorted_data+j+1)=key;
	}
}

/* Sort data using bubble sort.*/
static void bubble_sort(long **unsorted_data_ptr, long n, int ascending)
{
	long temp=0;
	long idx_in=0;
	long idx_out=0;
	long *unsorted_data = *unsorted_data_ptr;

	if (ascending ^ 1)
	{
		for (idx_out=0; idx_out<n; idx_out++)
		{
			for (idx_in=0; idx_in<n-idx_out-1; idx_in++)
			{
				if (*(unsorted_data+idx_in)>*(unsorted_data+idx_in+1))
				{
					temp= *(unsorted_data+idx_in);
					*(unsorted_data+idx_in)=*(unsorted_data+idx_in+1);
					*(unsorted_data+idx_in+1)=temp;
				}
			}
		}
	}
	else
	{
		for (idx_out=0; idx_out<n; idx_out++)
		{
			for (idx_in=0; idx_in<n-idx_out-1; idx_in++)
			{
				if (*(unsorted_data+idx_in)<*(unsorted_data+idx_in+1))
				{
					temp= *(unsorted_data+idx_in);
					*(unsorted_data+idx_in)=*(unsorted_data+idx_in+1);
					*(unsorted_data+idx_in+1)=temp;
				}
			}
		}
	}
}

/* Sort data using quick sort.*/
static void quick_sort(long **unsorted_data_ptr, long lb, long ub, int ascending)
{
	long *pivot_position=(long*)malloc(sizeof(long));

	if (lb>=ub)
		return;
	
	partition(unsorted_data_ptr,lb,ub,pivot_position,ascending);
	quick_sort(unsorted_data_ptr,lb,*pivot_position-1,ascending);
	quick_sort(unsorted_data_ptr,*pivot_position+1,ub,ascending);
}

/* Function to partition the data in quicksort. */
static void partition(long **unsorted_data_ptr, long lb, long ub, long *pivot_position, int ascending)
{
	long up=0L;
	long down=0L;
	long temp=0L;
	long pivot_elt=0L;
	long pivot_elt_index=0L;
	long *unsorted_data = *unsorted_data_ptr;

	up=ub;
	down=lb;
	pivot_elt_index= (lb+ub)/2;
	pivot_elt = *(unsorted_data+pivot_elt_index);

	while (down < up)
	{
		while (*(unsorted_data+down)<=pivot_elt && down < ub)
			down++;
		
		while (*(unsorted_data+up)>pivot_elt)
			up--;

		if (down < up)
		{
			if (down==pivot_elt_index)
			       pivot_elt_index=up;
			if (up==pivot_elt_index)
			       pivot_elt_index=down;
			temp = *(unsorted_data+down);
			*(unsorted_data+down)=*(unsorted_data+up);
			*(unsorted_data+up)=temp;
		}
	}
	*(unsorted_data+pivot_elt_index)=*(unsorted_data+up);
	*(unsorted_data+up)=pivot_elt;
	*pivot_position=up;
}
