/***************************************************************************
 *   Copyright (C) 2006 by Parthasarathy   *
 *   sarathy@Zion   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>

#include <input.h>
#include <output.h>
#include <algorithms.h>
#include <constants.h>
#include <test_algorithms.h>

int main(int argc, char *argv[])
{
	long n=0;
	int help=0;
	int order=0;
	int option=0;
	int nosort=0;
	
	int verify_result=0;
	int new_input=0;
	int algorithm=0;
	int compressor=0;

	long *data=NULL;	
	char *input_file=(char*) malloc (MAX_FILENAME_LENGTH);
	char *output_file=(char*) malloc (MAX_FILENAME_LENGTH);
	char *stats_file=(char*) malloc (MAX_FILENAME_LENGTH);

        struct timeval sort_start;
        struct timezone tzone_start;

        struct timeval sort_end;
        struct timezone tzone_end;

        double start=0.0F;
        double end=0.0F;

        FILE *sort_graph_data;

	/* Dont display system error messages for options.*/
	opterr=0;

	/* Get all the provided option values.*/
	while ((option=getopt(argc,argv,"i:o:n:ca:r:s:hd:xv"))!=-1)
	{
		switch (option)
		{
			/* Input data file containing unsorted data.*/
			case 'i': input_file=optarg;
				  break;
				  
			/* Output data file containing sorted data.*/		
			case 'o': output_file=optarg;
				  break;

			/* Sorting statistics file.*/		
			case 's': stats_file=optarg;
				  break;

			/* Number of elements to be sorted.*/
			case 'n': n = atol(optarg);
				  break;

			/* Sorting order 0-ascending, 1-descending.*/
			case 'd': order = atoi(optarg);
				  break;

			/* Create new data rather than operating on old data. */
			case 'c': new_input = 1;
				  break;

			/* Sorting algorithm to be used. */
			case 'a': algorithm = atoi(optarg);
				  break;

			/* To reduce the size of input data since rand() produces numbers in the range of 0 - 2^31.*/
			case 'r': compressor = atoi(optarg);
				  break;

			/* To display display_help */
			case 'h': help=1;
				  break;

			/* Dont sort. Just create the data in the specified input file.*/
			case 'x': nosort=1;
				  break;
				  
			/* Verify whether the output is sorted or not.*/
			case 'v': verify_result=1;
				  break;

			/* Unknown option. */
			case '?': 
				  if (isprint(optopt))
					  printf ("Unknown option -%c\n",optopt);
				  else
					  printf ("Unknown option \\x%x\n",optopt);
				  return 1;

			default:
				  abort();
		}
	}

	/* Validation of input. */
	if (n<=0 || strlen(input_file)==0 || (algorithm<=0&&nosort==0))
	{
		display_help();	
		return 1;
	}

	/* Allocate memory depending upon input size. */	
	data= (long *) malloc (n*sizeof(long));

	/* If new data is needed.*/	
	if (new_input&&1 || nosort&&1)
		create_data(input_file ,n,DELIMITER_SPACE,compressor);

	/* Do not sort. Just create new data and load it in the specified i/p file.*/
	if (nosort&&1)
		exit(0);

	/* Load the data from file to memory. */
	load_data(input_file ,n,DELIMITER_SPACE,&data);

	/* Sort */	
	switch (algorithm)
	{
		case INSERTION_SORT:
			gettimeofday(&sort_start ,&tzone_start);
			sort_data (&data, n, INSERTION_SORT, order);
			gettimeofday(&sort_end,&tzone_end);
			break;
			
		case SELECTION_SORT:
			gettimeofday(&sort_start ,&tzone_start);
			sort_data (&data, n, SELECTION_SORT, order);
			gettimeofday(&sort_end,&tzone_end);
			break;

		case MERGE_SORT:
			gettimeofday(&sort_start ,&tzone_start);
			sort_data (&data, n, MERGE_SORT, order);
			gettimeofday(&sort_end,&tzone_end);
			break;

		case INSERTION_RECURSIVE_SORT:
			gettimeofday(&sort_start ,&tzone_start);
			sort_data (&data, n, INSERTION_RECURSIVE_SORT, order);
			gettimeofday(&sort_end,&tzone_end);
			break;

		case BUBBLE_SORT:
			gettimeofday(&sort_start ,&tzone_start);
			sort_data(&data, n, BUBBLE_SORT, order);
			gettimeofday(&sort_end,&tzone_end);
			break;

		case QUICK_SORT:
			gettimeofday(&sort_start ,&tzone_start);
			sort_data(&data, n, QUICK_SORT, order);
			gettimeofday(&sort_end,&tzone_end);
			break;

		default:
			abort();
	}

	/* Output sorted data if needed. */	
	if (strlen(output_file)>0)
	{
		if (strcmp(output_file,STDOUT)==0)
			console_output(&data,n);
		else
			file_output(output_file,&data,n);
	}

	/* Output sorting statistics if needed.*/
	if (strlen(stats_file)>0)
	{
		start= (float) sort_start.tv_sec + (float)sort_start.tv_usec/1000000.0F;
        	end= (float) sort_end.tv_sec + (float)sort_end.tv_usec/1000000.0F;
		if (strcmp(output_file,STDOUT)==0)
			fprintf (stdout,"%ld %lf\n",n,end-start);
		else
		{
			sort_graph_data=fopen(stats_file,"a+");
			fprintf (sort_graph_data,"%ld %lf\n",n,end-start);
			fclose(sort_graph_data);
		}
	}

	/* Verify the results if needed.*/
	if (verify_result&&1)
		verify(data,n,order);
	
	/* Free the memory used for storing data. */
	free(data);
	return 0;
}
